/*  tx.c

    Source file for  tx  (data structure transmitter module).

    Copyright (C) 1993-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will read in a Karma data file and transmit the
    multi-array over one or more Karma connections to a Karma server module.
    The module will also allow a list of files to be transmitted, one after the
    other.


    Written by      Richard Gooch   4-OCT-1992

    Updated by      Richard Gooch   19-DEC-1992

    Updated by      Richard Gooch   1-JAN-1993: Took account of change to
  ch_gets  .

    Updated by      Richard Gooch   27-MAR-1993: Took account of change to
  dsxfr_get_multi  .

    Updated by      Richard Gooch   17-MAY-1993: Took account of change to
  dsxfr_get_multi  .

    Updated by      Richard Gooch   24-JUL-1993: Changed to  ch_open_file  .

    Updated by      Richard Gooch   8-AUG-1993: Tidied up by deallocating
  multi_array data structure when finished.

    Updated by      Richard Gooch   5-OCT-1993: Changed over to  panel_
  package for command line user interface.

    Updated by      Richard Gooch   6-OCT-1993: Moved  main  into this file.

    Updated by      Richard Gooch   23-NOV-1993: Changed to use of
  ex_word_skip  .

    Updated by      Richard Gooch   10-MAR-1995: Added support for "spray"
  protocol.

    Updated by      Richard Gooch   14-JUN-1995: Made use of <ex_uint>.

    Updated by      Richard Gooch   9-AUG-1995: Added support for "hog_request"
  protocol.

    Updated by      Richard Gooch   29-JAN-1996: Added call to
  <im_register_lib_version>.

    Updated by      Richard Gooch   1-JUN-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-AUG-1996: Upgraded "spray" protocol to
  support synchronisation.

    Updated by      Richard Gooch   23-MAY-1997: Added "ping_server" protocol.

    Updated by      Richard Gooch   3-SEP-1997: Switched to ERRSTRING.

    Updated by      Richard Gooch   1-OCT-1997: Added "server_exit" protocol.

    Updated by      Richard Gooch   27-DEC-1997: Added "udp_spray" protocol.

    Updated by      Richard Gooch   28-DEC-1997: Added spray pattern and
  maximum transmit rate.

    Updated by      Richard Gooch   4-FEB-1998: Bug fix in <udp_spray> when
  insufficient arguments given.

    Updated by      Richard Gooch   18-FEB-1997: Removed "udp_spray" protocol
  and added datagram support for "spray" protocol.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   13-DEC-1998: Added call to <sched_yield>
  in <delay>.

    Updated by      Richard Gooch   2-FEB-1999: Disabled call to <sched_yield>
  for OS_Solaris: dumbfucks require -lposix4.

    Updated by      Richard Gooch   19-FEB-1999: Switched back from <write> to
  <ch_write> in <datagram_spray> as XTI interface doesn't support <write>.

    Updated by      Richard Gooch   2-MAR-1999: Added call to <ch_flush> in
  <datagram_spray> to make packets < sdu_max work.

    Updated by      Richard Gooch   25-AUG-1999: Added #ifdef for IRIX 6.2.

    Last updated by Richard Gooch   7-OCT-2000: Took account of interface
  change to function <dm_native_wait_input>.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#if !defined(OS_SunOS) && !defined(OS_IRIX6)
#  include <sched.h>
#endif
#include <sys/time.h>
#include <sys/times.h>
#include <sys/file.h>
#include <errno.h>
#include <karma.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_conn.h>
#include <karma_pio.h>
#include <karma_ex.h>
#include <karma_ds.h>
#include <karma_dm.h>
#include <karma_im.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_s.h>
#include <karma_m.h>
#include <karma_n.h>
#include <karma_r.h>


#define VERSION "1.7"
#define BUF_SIZE 262144
#define DEFAULT_DGRAM_SIZE 4096


STATIC_FUNCTION (flag tx, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (flag transmit_arrayfile, (char *arrayfile) );
STATIC_FUNCTION (void run_animate_file, (char *animate_file) );
STATIC_FUNCTION (char *get_spray_buffer, (unsigned int size) );
STATIC_FUNCTION (void do_spray, (CONST char *p) );
STATIC_FUNCTION (void ping, (CONST char *p) );
STATIC_FUNCTION (void datagram_spray,
		 (Channel channel, Channel datagram_channel,
		  unsigned int packet_size, unsigned int num_packets,
		  char *buffer) );
STATIC_FUNCTION (void delay, (unsigned long delay_usec) );


/*  Private data  */
static flag animate_flag = FALSE;
static int num_iterations = 1;
static flag synchronous_spray = TRUE;
static unsigned long transmit_rate = 0;

#define SPRAY_DATA_PATTERN     0
#define SPRAY_DATA_RANDOM      1
#define NUM_SPRAY_DATA_CHOICES 2
static char *spray_data_choices[NUM_SPRAY_DATA_CHOICES] =
{
    "pattern",
    "random",
};
static unsigned int spray_type = SPRAY_DATA_PATTERN;

#define MAX_PATTERN_LENGTH 16
static unsigned char spray_pattern[MAX_PATTERN_LENGTH] = {0};
static unsigned int pattern_length = 1;


int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    conn_register_client_protocol ("spray", 2, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("hog_request", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("ping_server", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("server_exit", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "transmit_rate","maximum transmit rate (bytes/sec)",
		    K_ULONG, (char *) &transmit_rate,
		    PIA_END);
    panel_add_item (panel, "do_spray", "bytes [packet_size]",
		    PIT_FUNCTION, (void *) do_spray,
		    PIA_END);
    panel_add_item (panel, "synchronous_spray", "wait for sync at end",
		    PIT_FLAG,
		    &synchronous_spray,
		    PIA_END);
    panel_add_item (panel, "spray_data", "choice of spray data",
		    PIT_CHOICE_INDEX, (char *) &spray_type,
		    PIA_NUM_CHOICE_STRINGS, NUM_SPRAY_DATA_CHOICES,
		    PIA_CHOICE_STRINGS, spray_data_choices,
		    PIA_END);
    panel_add_item (panel, "ping", "ping specified number of times",
		    PIT_FUNCTION, (void *) ping,
		    PIA_END);
    panel_add_item (panel, "pattern", "spray pattern", K_UBYTE, spray_pattern,
		    PIA_ARRAY_MIN_LENGTH, 1,
		    PIA_ARRAY_MAX_LENGTH, MAX_PATTERN_LENGTH,
		    PIA_ARRAY_LENGTH, (char *) &pattern_length,
		    PIA_END);
    panel_add_item (panel, "num_iterations", "number", K_INT, &num_iterations,
		    PIA_END);
    panel_add_item (panel, "animate", "flag", PIT_FLAG, &animate_flag,
		    PIA_END);
    panel_push_onto_stack (panel);
    im_register_lib_version (KARMA_VERSION);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile");
    module_run (argc, argv, "tx", VERSION, tx, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag tx (CONST char *command, FILE *fp)
{
    char filename[STRING_LENGTH+1];
    extern int animate_flag;

    for ( ; command != NULL; command = ex_word_skip (command) )
    {
	sscanf (command, "%s", filename);
	if (animate_flag) run_animate_file (filename);
	else transmit_arrayfile (filename);
    }
    return (TRUE);
}   /*  End Function tx  */

static flag transmit_arrayfile (char *arrayfile)
/*  This routine will transmit an arrayfile over any Karma "multi_array"
    connections.
    The name of the arrayfile must be pointed to by  arrayfile  .
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    multi_array *multi_desc;
    static char function_name[] = "operate_file";

    if (arrayfile == NULL)
    {
	fprintf (stderr, "NULL array pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Load input array  */
    if ( ( multi_desc = dsxfr_get_multi (arrayfile, TRUE, K_CH_MAP_LOCAL,
					 FALSE) )
	== NULL )
    {
	return (FALSE);
    }
    /*  Transmit multi-array over connection  */
    if (dsxfr_put_multi ("connections", multi_desc) != TRUE)
    {
	fprintf (stderr, "Error transmitting data over connection\n");
	/*  Decrement attachment count  */
	ds_dealloc_multi (multi_desc);
	return (FALSE);
    }
    /*  Decrement attachment count  */
    ds_dealloc_multi (multi_desc);
    return (TRUE);
}   /*  End Function transmit_arrayfile  */

static void run_animate_file (char *animate_file)
/*  This routine will read and transmit a number of arrayfiles over any Karma
    "multi_array" connections.
    The names of the arrayfiles to transmit must be contained in the file
    with name pointed to by  animate_file  .
    The routine returns nothing.
*/
{
    Channel channel;
    int iter_count;
    char line[STRING_LENGTH + 1];
    extern int num_iterations;

    for (iter_count = 0; (iter_count < num_iterations) || (num_iterations < 0);
	 ++iter_count)
    {
	/*  Load animate file  */
	if ( ( channel = ch_open_file (animate_file, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     animate_file, ERRSTRING);
	    return;
	}
	/*  Read lines  */
	while ( ch_gets (channel, line, STRING_LENGTH) )
	{
	    /*  Not end of file  */
	    if ( (line[0] == '#') || (line[0] == '\n') )
	    {
		/*  Comment or blank line: skip  */
		continue;
	    }
	    if ( !transmit_arrayfile (line) ) return;
	    if ( s_check_for_int () )
	    {
		fprintf (stderr,
			 "Control C: quitting animate file: \"%s\" after: %d iterations\n",
			 animate_file, iter_count);
		ch_close (channel);
		return;
	    }
	}
	ch_close (channel);
    }
}   /*  End Function run_animate_file  */

static char *get_spray_buffer (unsigned int size)
/*  [SUMMARY] Get the current spray buffer.
    <size> The size of the buffer.
    [RETURNS] A pointer to a filled spray buffer.
*/
{
    unsigned int byte_count;
    static unsigned int rbuf_len = 0;
    static unsigned int vbuf_len = 0;
    static char *rbuffer = NULL;
    static char *vbuffer = NULL;
    static char function_name[] = "get_spray_buffer";

    if ( (spray_type == SPRAY_DATA_RANDOM) && (rbuf_len < size) )
    {
	if (rbuffer != NULL) m_free (rbuffer);
	if ( ( rbuffer = m_alloc (size) ) == NULL )
	{
	    m_abort (function_name, "random number buffer");
	}
	fprintf (stderr, "Filling buffer with random numbers...");
	for (byte_count = 0; byte_count < size; ++byte_count)
	{
	    *(unsigned char *) (rbuffer + byte_count) = n_uniform () * 255.0;
	}
	rbuf_len = size;
	fprintf (stderr, "done\n");
    }
    if ( (spray_type != SPRAY_DATA_RANDOM) && (vbuf_len < size) )
    {
	if (vbuffer != NULL) m_free (vbuffer);
	if ( ( vbuffer = m_alloc (size) ) == NULL )
	{
	    m_abort (function_name, "value buffer");
	}
	vbuf_len = size;
    }
    switch (spray_type)
    {
      case SPRAY_DATA_PATTERN:
	byte_count = size / pattern_length;
	m_fill (vbuffer, pattern_length, (CONST char *) spray_pattern,
		pattern_length, byte_count);
	m_copy (vbuffer + byte_count * pattern_length,
		(CONST char *) spray_pattern,
		size - byte_count * pattern_length);
	break;
      case SPRAY_DATA_RANDOM:
	break;
      default:
	fprintf (stderr, "Illegal spray type: %u\n", spray_type);
	a_prog_bug (function_name);
	break;
    }
    return (spray_type == SPRAY_DATA_RANDOM) ? rbuffer : vbuffer;
}   /*  End Function get_spray_buffer  */

static void do_spray (CONST char *p)
/*  [PURPOSE] This routine will spray bytes to all "spray" servers.
    <p> A string containing the number of bytes to spray.
    [RETURNS] Nothing.
*/
{
    Connection connection;
    Channel channel, datagram_channel;
    flag wrote_all;
    char dummy;
    unsigned int num_bytes, packet_size, size;
    unsigned int byte_count, block_size, conn_count, num_conn;
    unsigned long interval;
    float time_taken, usec, rate;
    struct karma_qos qos;
    struct timeval start_time;
    struct timeval stop_time;
    char *name, *buffer;
    extern flag synchronous_spray;
    static struct timezone tz = {0, 0};
    static char function_name[] = "do_spray";

    if (p == NULL)
    {
	fprintf (stderr, "Must specify number of bytes to spray\n");
	return;
    }
    if ( ( num_bytes = ex_uint (p, &p) ) < 1 )
    {
	fprintf (stderr, "Must specify at least 1 byte\n");
	return;
    }
    packet_size = (p == NULL) ? BUF_SIZE : ex_uint (p, &p);
    if ( ( num_conn = conn_get_num_client_connections ("spray") ) < 1 )
    {
	fprintf (stderr, "No connections to spray servers\n");
	return;
    }
    buffer = get_spray_buffer (packet_size);
    if (transmit_rate < 1) interval = 0;
    else interval = ( (float) packet_size / (float) transmit_rate * 1e6 );
    for (conn_count = 0; conn_count < num_conn; ++conn_count)
    {
	if ( ( connection = conn_get_client_connection ("spray", conn_count) )
	    == NULL )
	{
	    fprintf (stderr, "Connection: %u does not exist!\n", conn_count);
	    a_prog_bug (function_name);
	}
	if ( ( name = conn_get_connection_module_name (connection) ) == NULL )
	{
	    fprintf (stderr, "NULL module name for connection: %u\n",
		     conn_count);
	    a_prog_bug (function_name);
	}
	channel = conn_get_channel (connection);
	datagram_channel = conn_get_datagram_channel (connection);
	fprintf (stderr, "Spraying connection: %u (module: \"%s\")...   ",
		 conn_count, name);
	if (datagram_channel != NULL)
	{
	    ch_get_qos (datagram_channel, &qos);
	    size = (packet_size <qos.tx.sdu_max) ? packet_size :qos.tx.sdu_max;
	    datagram_spray (channel, datagram_channel, size, num_bytes / size,
			    buffer);
	    continue;
	}
	if (gettimeofday (&start_time, &tz) != 0)
	{
	    fprintf (stderr, "Error getting time of day\t%s%c\n",
		     ERRSTRING, BEL);
	    exit (RV_SYS_ERROR);
	}
	/*  Tell server how many bytes to expect  */
	if ( !pio_write32 (channel, num_bytes) ) continue;
	/*  Tell server whether synchronous operation is required  */
	if ( !pio_write32 (channel, synchronous_spray) ) continue;
	/*  Write all the bytes  */
	byte_count = num_bytes;
	wrote_all = TRUE;
	while (byte_count > 0)
	{
	    block_size = (byte_count > packet_size) ? packet_size : byte_count;
	    if (ch_write (channel, buffer, block_size) < block_size)
	    {
		fprintf (stderr, "Error writing data\t%s\n", ERRSTRING);
		byte_count = 0;
		wrote_all = FALSE;
		break;
	    }
	    byte_count -= block_size;
	    if ( !ch_flush (channel) )
	    {
		fprintf (stderr, "Error flushing\n");
		wrote_all = FALSE;
		break;
	    }
	    delay (interval);
	}
	if (!wrote_all) continue;
	if (synchronous_spray)
	{
	    /*  Read the return byte  */
	    if (ch_read (channel, &dummy, 1) < 1) continue;
	}
	if (gettimeofday (&stop_time, &tz) != 0)
	{
	    fprintf (stderr, "Error getting time of day\t%s%c\n",
		     ERRSTRING, BEL);
	    exit (RV_SYS_ERROR);
	}
	usec = (float) (stop_time.tv_usec - start_time.tv_usec) * 1e-6;
	time_taken = (float) (stop_time.tv_sec - start_time.tv_sec) + usec;
	rate = (float) num_bytes / time_taken;
	fprintf (stderr, "%u bytes/second\n", (unsigned int) rate);
    }
}   /*  End Function do_spray  */

static void ping (CONST char *p)
/*  [PURPOSE] This routine will ping all "ping_server" servers.
    <p> A string containing the number of times to ping.
    [RETURNS] Nothing.
*/
{
    Connection connection;
    unsigned int num_pings, size;
    unsigned int conn_count, num_conn, ping_count;
    signed long tx_usec, rtt_usec, rx_usec;
    signed long min_rtt_time, av_rtt_time, max_rtt_time;
    signed long min_tx_time, av_tx_time, max_tx_time;
    signed long min_rx_time, av_rx_time, max_rx_time;
    struct timeval tx_start_time, tx_done_time, rx_start_time, rx_done_time;
    char *name;
    char buffer[BUF_SIZE];
    static char function_name[] = "ping";

    if (p == NULL)
    {
	fprintf (stderr, "Must specify number of times to ping\n");
	return;
    }
    if ( ( num_pings = ex_uint (p, &p) ) < 1 )
    {
	fprintf (stderr, "Must specify at least 1 ping\n");
	return;
    }
    if (p == NULL) size = 1;
    else size = ex_uint (p, &p);
    if (size > BUF_SIZE)
    {
	fprintf (stderr, "Size: %u too large, %u bytes maximum\n",
		 size, BUF_SIZE);
	return;
    }
    m_clear (buffer, size);
    if ( ( num_conn = conn_get_num_client_connections ("ping_server") ) < 1 )
    {
	fprintf (stderr, "No connections to ping servers\n");
	return;
    }
    fprintf (stderr, "Pinging each ping server %d times with size %u bytes\n",
	     num_pings, size);
    for (conn_count = 0; conn_count < num_conn; ++conn_count)
    {
	Channel channel;
	int fd;

	if ( ( connection = conn_get_client_connection ("ping_server",
							conn_count) ) == NULL )
	{
	    fprintf (stderr, "Connection: %u does not exist!\n", conn_count);
	    a_prog_bug (function_name);
	}
	if ( ( name = conn_get_connection_module_name (connection) ) == NULL )
	{
	    fprintf (stderr, "NULL module name for connection: %u\n",
		     conn_count);
	    a_prog_bug (function_name);
	}
	channel = conn_get_datagram_channel (connection);
	if (channel == NULL) channel = conn_get_channel (connection);
	fd = ch_get_descriptor (channel);
	min_rtt_time = 1000000;
	max_rtt_time = 0.0;
	av_rtt_time = 0.0;
	min_tx_time = 1000000;
	max_tx_time = 0.0;
	av_tx_time = 0.0;
	min_rx_time = 1000000;
	max_rx_time = 0.0;
	av_rx_time = 0.0;
	fprintf (stderr, "Pinging connection: %u (module: \"%s\")...\n",
		 conn_count, name);
	fprintf (stderr, "Iter\ttxt (us)    rtt (us)    rxt (us)\n");
	for (ping_count = 0; ping_count < num_pings; ++ping_count)
	{
	    if (gettimeofday (&tx_start_time, NULL) != 0)
	    {
		fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    if (ch_write (channel, buffer, size) < size)
	    {
		
		fprintf (stderr, "Error writing\t%s\n", ERRSTRING);
		continue;
	    }
	    if ( !ch_flush (channel) )
	    {
		
		fprintf (stderr, "Error flushing\t%s\n", ERRSTRING);
		continue;
	    }
	    if (gettimeofday (&tx_done_time, NULL) != 0)
	    {
		fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    dm_native_wait_input (fd, -1);
	    if (gettimeofday (&rx_start_time, NULL) != 0)
	    {
		fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    if (ch_read (channel, buffer, size) < size)
	    {
		
		fprintf (stderr, "Error reading\t%s\n", ERRSTRING);
		continue;
	    }
	    if (gettimeofday (&rx_done_time, NULL) != 0)
	    {
		fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    tx_usec = tx_done_time.tv_usec - tx_start_time.tv_usec;
	    tx_usec += (tx_done_time.tv_sec - tx_start_time.tv_sec) * 1000000;
	    if (tx_usec < min_tx_time) min_tx_time = tx_usec;
	    if (tx_usec > max_tx_time) max_tx_time = tx_usec;
	    av_tx_time += tx_usec;
	    rtt_usec = rx_start_time.tv_usec - tx_done_time.tv_usec;
	    rtt_usec += (rx_start_time.tv_sec - tx_done_time.tv_sec) * 1000000;
	    if (rtt_usec < min_rtt_time) min_rtt_time = rtt_usec;
	    if (rtt_usec > max_rtt_time) max_rtt_time = rtt_usec;
	    av_rtt_time += rtt_usec;
	    rx_usec = rx_done_time.tv_usec - rx_start_time.tv_usec;
	    rx_usec += (rx_done_time.tv_sec - rx_start_time.tv_sec) * 1000000;
	    if (rx_usec < min_rx_time) min_rx_time = rx_usec;
	    if (rx_usec > max_rx_time) max_rx_time = rx_usec;
	    av_rx_time += rx_usec;
#if 0
	    printf ("%u\t%-10ld  %-10ld  %ld\n",
		    ping_count, tx_usec, rtt_usec, rx_usec);
#endif
	    usleep (50000);
	}
	fflush (stdout);
	av_tx_time /= num_pings;
	av_rtt_time /= num_pings;
	av_rx_time /= num_pings;
	fprintf (stderr, "Min:\t%-10ld  %-10ld  %ld\n",
		 min_tx_time, min_rtt_time, min_rx_time);
	fprintf (stderr, "Av:\t%-10ld  %-10ld  %ld\n",
		 av_tx_time, av_rtt_time, av_rx_time);
	fprintf (stderr, "Max:\t%-10ld  %-10ld  %ld\n",
		 max_tx_time, max_rtt_time, max_rx_time);
    }
}   /*  End Function ping  */

static void datagram_spray (Channel channel, Channel datagram_channel,
			    unsigned int packet_size, unsigned int num_packets,
			    char *buffer)
/*  [PURPOSE] This routine will spray bytes to a datagram channel.
    <channel> The stream channel.
    <datagram_channel> The datagram channel.
    <packet_size> The size of the packets.
    <num_packets> The number of packets to send.
    <buffer> The data to send.
    [RETURNS] Nothing.
*/
{
    unsigned long num_packets_received, num_dropped, num_bytes_received;
    unsigned long num_partial_packets, interval;
    unsigned int packet_count;
    unsigned int num_bytes = 0;
    float time_taken, usec, rate;
    struct timeval start_time;
    struct timeval stop_time;
    static struct timezone tz = {0, 0};
    static char function_name[] = "datagram_spray";

    if (transmit_rate < 1) interval = 0;
    else interval = ( (float) packet_size / (float) transmit_rate * 1e6 );
    if (num_packets < 1)
    {
	fprintf (stderr, "no whole packets to send\n");
	return;
    }
    if ( !pio_write32 (channel, packet_size) || !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error writing packet size\n", function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_packets_received) )
    {
	fprintf (stderr, "%s: error reading dummy num_packets_received\n",
		 function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_bytes_received) )
    {
	fprintf (stderr, "%s: error reading dummy num_bytes_received\n",
		 function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_partial_packets) )
    {
	fprintf (stderr, "%s: error reading dummy num_partial_packets\n",
		 function_name);
	return;
    }
    if (gettimeofday (&start_time, &tz) != 0)
    {
	    fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
    }
    /*  Write all the packets  */
    for (packet_count = 0; packet_count < num_packets; ++packet_count)
    {
	if (ch_write (datagram_channel, buffer, packet_size) < packet_size)
	{
	    fprintf (stderr, "%s: error writing data\t%s\n",
		     function_name, ERRSTRING);
	    return;
	}
	if ( !ch_flush (datagram_channel) )
	{
	    fprintf (stderr, "%s: error flusing data\t%s\n",
		     function_name, ERRSTRING);
	    return;
	}
	num_bytes += packet_size;
	delay (interval);
    }
    if ( !pio_write32 (channel, 0) ) return;
    if ( !ch_flush (channel) ) return;
    if ( !pio_read32 (channel, &num_packets_received) )
    {
	fprintf (stderr, "%s: error reading num_packets_received\n",
		 function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_bytes_received) )
    {
	fprintf (stderr, "%s: error reading num_bytes_received\n",
		 function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_partial_packets) )
    {
	fprintf (stderr, "%s: error reading num_partial_packets\n",
		 function_name);
	return;
    }
    if (gettimeofday (&stop_time, &tz) != 0)
    {
	fprintf (stderr, "Error getting time of day\t%s%c\n",
		 ERRSTRING, BEL);
	exit (RV_SYS_ERROR);
    }
    usec = (float) (stop_time.tv_usec - start_time.tv_usec) * 1e-6;
    time_taken = (float) (stop_time.tv_sec - start_time.tv_sec) + usec;
    rate = (float) num_bytes / time_taken;
    fprintf (stderr, "%u bytes/second transmitted\n", (unsigned int) rate);
    num_dropped = num_packets - num_packets_received;
    fprintf (stderr, "  %lu packets dropped, %u sent (%lu%% loss) ",
	     num_dropped, num_packets,
	     num_dropped * 100 / num_packets);
    rate = (float) num_bytes_received / time_taken;
    fprintf (stderr, "%u bytes/second received\n", (unsigned int) rate);
    fprintf (stderr, "  %lu partial packets received\n", num_partial_packets);
}   /*  End Function datagram_spray  */

static void delay (unsigned long delay_usec)
/*  [SUMMARY] Wait a specified amount of time.
    <delay_usec> The number of microseconds to wait. If this is zero the
    routine does not wait.
    [NOTE] This routine depends on the gettimeofday system call.
    [RETURNS] Nothing.
*/
{
    signed long time_taken;
    unsigned long delay_sec, delay_remainder;
    struct timeval start_time;
    struct timeval stop_time;

    if (delay_usec < 1) return;
    if (gettimeofday (&start_time, NULL) != 0)
    {
	fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    delay_sec = delay_usec / 1000000;
    delay_remainder = delay_usec - delay_sec * 1000000;
    while (TRUE)
    {
	if (gettimeofday (&stop_time, NULL) != 0)
	{
	    fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (stop_time.tv_sec == start_time.tv_sec)
	{
	    if (stop_time.tv_usec >= start_time.tv_usec + delay_usec) return;
#if !defined(OS_Solaris) && !defined(OS_SunOS) && !defined(OS_IRIX6)
	    sched_yield ();
#endif
	    continue;
	}
	/*  In a new second  */
	stop_time.tv_sec -= delay_sec;
	time_taken = (signed long) stop_time.tv_sec;
	time_taken -= (signed long) start_time.tv_sec;
	time_taken *= 1000000;
	time_taken += (signed long) stop_time.tv_usec;
	time_taken -= (signed long) start_time.tv_usec;
	if (time_taken >= (signed long) delay_remainder) return;
#if !defined(OS_Solaris) && !defined(OS_SunOS) && !defined(OS_IRIX6)
	sched_yield ();
#endif
    }
}   /*  End Function delay  */
